@@ -4,41 +4,54 @@ use 5.008000;
 use strict;
 use warnings;
 
-our $VERSION = '0.08';
+our $VERSION = '0.09';
 
 require Exporter;
 
 our @ISA = qw(Exporter);
 
 our @EXPORT = qw(
-    box_keypair
-    sign_keypair
-    crypto_sign
+    real_crypto_stream_xor
     real_crypto_sign
-    crypto_sign_open
     real_crypto_sign_open
+    real_crypto_box
+    real_crypto_hash
+    real_crypto_box_open
+    real_crypto_secretbox
+    real_crypto_secretbox_open
+    real_crypto_stream
+
+    crypto_sign
+    crypto_sign_open
     crypto_sign_detached
     crypto_sign_verify_detached
     crypto_box
-    real_crypto_box
     crypto_box_open
-    real_crypto_box_open
     crypto_secretbox
-    real_crypto_secretbox
     crypto_secretbox_open
-    real_crypto_secretbox_open
     crypto_hash
-    real_crypto_hash
     crypto_generichash
     crypto_generichash_key
     crypto_stream
-    real_crypto_stream
     crypto_stream_xor
-    real_crypto_stream_xor
     crypto_box_keypair
     crypto_sign_keypair
+    crypto_pwhash_scrypt_str
+    crypto_pwhash_scrypt_str_verify
+    crypto_box_nonce
+    crypto_stream_key
+    crypto_stream_nonce
+    crypto_pwhash_salt
+    crypto_pwhash_scrypt
+    crypto_scalarmult_base
+    crypto_scalarmult
+    crypto_scalarmult_safe
+
     randombytes_buf
     randombytes_random
+    box_keypair
+    sign_keypair
+
     crypto_stream_NONCEBYTES
     crypto_stream_KEYBYTES
     crypto_box_NONCEBYTES
@@ -50,23 +63,20 @@ our @EXPORT = qw(
     crypto_sign_SECRETKEYBYTES
     crypto_sign_BYTES
     crypto_secretbox_MACBYTES
-    crypto_stream_key
-    crypto_stream_nonce
-    crypto_pwhash_salt
-    crypto_pwhash_scrypt
+    crypto_secretbox_KEYBYTES
+    crypto_secretbox_NONCEBYTES
     crypto_pwhash_SALTBYTES
     crypto_pwhash_OPSLIMIT
     crypto_pwhash_MEMLIMIT
     crypto_pwhash_STRBYTES
-    crypto_pwhash_scrypt_str
-    crypto_pwhash_scrypt_str_verify
-    crypto_box_nonce
     crypto_generichash_KEYBYTES
     crypto_generichash_KEYBYTES_MIN
     crypto_generichash_KEYBYTES_MAX
     crypto_generichash_BYTES
     crypto_generichash_BYTES_MIN
     crypto_generichash_BYTES_MAX
+    crypto_scalarmult_SCALARBYTES
+    crypto_scalarmult_BYTES
 );
 
 use subs qw/
@@ -77,7 +87,9 @@ use subs qw/
     crypto_box_SECRETKEYBYTES
     crypto_box_MACBYTES
     crypto_box_SEEDBYTES
+    crypto_secretbox_KEYBYTES
     crypto_secretbox_MACBYTES
+    crypto_secretbox_NONCEBYTES
     crypto_sign_PUBLICKEYBYTES
     crypto_sign_SECRETKEYBYTES
     crypto_sign_BYTES
@@ -91,6 +103,8 @@ use subs qw/
     crypto_generichash_BYTES
     crypto_generichash_BYTES_MIN
     crypto_generichash_BYTES_MAX
+    crypto_scalarmult_SCALARBYTES
+    crypto_scalarmult_BYTES
 /;
 
 require XSLoader;
@@ -108,6 +122,35 @@ sub crypto_stream_nonce {
     return randombytes_buf(crypto_stream_NONCEBYTES);
 }
 
+sub crypto_scalarmult_base {
+    my ($n) = @_;
+    unless (length($n) == crypto_scalarmult_SCALARBYTES) {
+        die "[fatal]: secret key must be exactly " . crypto_scalarmult_SCALARBYTES . " bytes long\n";
+    }
+    
+    return real_crypto_scalarmult_base($n);
+}
+
+sub crypto_scalarmult {
+    my ($n, $p) = @_;
+
+    unless (length($n) == crypto_scalarmult_SCALARBYTES) {
+        die "[fatal]: secret key must be exactly " . crypto_scalarmult_SCALARBYTES . " bytes long\n";
+    }
+    
+    unless (length($p) == crypto_scalarmult_BYTES) {
+        die "[fatal]: public key must be exactly " . crypto_scalarmult_BYTES . " bytes long\n";
+    }
+    
+    return real_crypto_scalarmult($n, $p);
+}
+
+sub crypto_scalarmult_safe {
+    my ($n, $p, $p2) = @_;
+    
+    return crypto_hash(crypto_scalarmult($n, $p) ^ $p2 ^ $p);
+}
+
 sub crypto_stream {
     my ($len, $n, $k) = @_;
 
@@ -249,8 +292,12 @@ sub crypto_box {
 sub crypto_secretbox_open {
     my ($c, $n, $sk) = @_;
 
-    unless (length($sk) == crypto_box_SECRETKEYBYTES) {
-        die "[fatal]: secret key must be exactly " . crypto_box_SECRETKEYBYTES . " bytes long.\n";
+    unless (length($n) == crypto_secretbox_NONCEBYTES) {
+        die "[fatal]: secretbox nonce must be exactly " . crypto_secretbox_NONCEBYTES . " bytes long.\n";
+    }
+
+    unless (length($sk) == crypto_secretbox_KEYBYTES) {
+        die "[fatal]: secretbox key must be exactly " . crypto_secretbox_KEYBYTES . " bytes long.\n";
     }
 
     return real_crypto_secretbox_open($c, length($c), $n, $sk);
@@ -259,8 +306,12 @@ sub crypto_secretbox_open {
 sub crypto_secretbox {
     my ($m, $n, $sk) = @_;
 
-    unless (length($sk) == crypto_box_SECRETKEYBYTES) {
-        die "[fatal]: secret key must be exactly " . crypto_box_SECRETKEYBYTES . " bytes long.\n";
+    unless (length($n) == crypto_secretbox_NONCEBYTES) {
+        die "[fatal]: secretbox nonce must be exactly " . crypto_secretbox_NONCEBYTES . " bytes long.\n";
+    }
+
+    unless (length($sk) == crypto_secretbox_KEYBYTES) {
+        die "[fatal]: secretbox key must be exactly " . crypto_secretbox_KEYBYTES . " bytes long.\n";
     }
 
     return real_crypto_secretbox($m, length($m), $n, $sk);
@@ -359,7 +410,8 @@ Crypt::Sodium - Perl bindings for libsodium (NaCL) https://github.com/jedisct1/l
 =item crypto_box($message, $nonce, $public_key, $secret_key)
    
    Usage: my $ciphertext = crypto_box($m, $n, $pk, $sk);
-   Note: $nonce must be at least crypto_box_NONCEBYTES long.
+   
+   Note:  $nonce must be at least crypto_box_NONCEBYTES long.
 
 =item crypto_box_open($ciphertext, $nonce, $public_key, $secret_key)
    
@@ -368,8 +420,9 @@ Crypt::Sodium - Perl bindings for libsodium (NaCL) https://github.com/jedisct1/l
 =item crypto_secretbox($message, $nonce, $key);
    
    Usage: my $ciphertext = crypto_secretbox($m, $n, $k);
-   Note: $nonce must be at least crypto_box_NONCEBYTES long,
-         $key must be at least crypto_box_SECRETKEYBYTES long.
+   
+   Note:  $nonce must be at least crypto_box_NONCEBYTES long,
+          $key must be at least crypto_box_SECRETKEYBYTES long.
 
 =item crypto_secretbox_open($ciphertext, $nonce, $key);
    
@@ -382,15 +435,17 @@ Crypt::Sodium - Perl bindings for libsodium (NaCL) https://github.com/jedisct1/l
 =item crypto_stream($length, $nonce, $key)
    
    Usage: my $stream = crypto_stream($length, $nonce, $key);
-   Note: $nonce must be at least crypto_stream_NONCEBYTES long, 
-         $key must be at least crypto_stream_KEYBYTES long.
+   
+   Note:  $nonce must be at least crypto_stream_NONCEBYTES long, 
+          $key must be at least crypto_stream_KEYBYTES long.
 
 =item crypto_stream_xor($message, $nonce, $key)
    
    Usage: my $ciphertext = crypto_stream_xor($message, $nonce, $key);
           my $cleartext = crypto_stream_xor($ciphertext, $nonce, $key);
-   Note: $nonce must be at least crypto_stream_NONCEBYTES long, 
-         $key must be at least crypto_stream_KEYBYTES long.
+          
+   Note:  $nonce must be at least crypto_stream_NONCEBYTES long, 
+          $key must be at least crypto_stream_KEYBYTES long.
 
 =item randombytes_buf($length)
    
@@ -415,20 +470,37 @@ Crypt::Sodium - Perl bindings for libsodium (NaCL) https://github.com/jedisct1/l
 =item crypto_pwhash_scrypt($password, $salt, $keylen, $opslimit, $memlimit)
    
    Usage: my $derivedkey = crypto_pwhash_scrypt($password, $salt, $keylen, $opslimit, $memlimit);
-   Note: $salt must be crypto_pwhash_SALTBYTES long, use crypto_pwhash_salt() to generate
-         $keylen maybe omitted, the default is crypto_box_SEEDBYTES
-         $opslimit maybe omitted, the default is crypto_pwhash_OPSLIMIT
-         $memlimit maybe omitted, the default is crypto_pwhash_MEMLIMIT
-         See L<http://doc.libsodium.org/password_hashing/README.html> for details>.
+   
+   Note:  $salt must be crypto_pwhash_SALTBYTES long, use crypto_pwhash_salt() to generate
+          $keylen maybe omitted, the default is crypto_box_SEEDBYTES
+          $opslimit maybe omitted, the default is crypto_pwhash_OPSLIMIT
+          $memlimit maybe omitted, the default is crypto_pwhash_MEMLIMIT
+          See L<http://doc.libsodium.org/password_hashing/README.html> for details>.
 
 =item crypto_pwhash_scrypt_str($password, $salt, $opslimit, $memlimit)
    
    Usage: my $hash_string = crypto_pwhash_scrypt_str($password, $salt);
-   Note: like the crypto_pwhash_scrypt function, this function can also take an opslimit and memlimit
-         value.  The default opslimit is exported into your namespace as crypto_pwhash_OPSLIMIT and the
-         default memlimit is exported as crypto_pwhash_MEMLIMIT, if you have a really important password
-         to hash and don't mind using 1GB of ram and 10s+ of CPU time on an i7-class CPU, you can use 
-         crypto_pwhash_OPSLIMIT_SENSITIVE and crypto_pwhash_MEMLIMIT_SENSITIVE instead.
+   
+   Note:  like the crypto_pwhash_scrypt function, this function can also take an opslimit and memlimit
+          value.  The default opslimit is exported into your namespace as crypto_pwhash_OPSLIMIT and the
+          default memlimit is exported as crypto_pwhash_MEMLIMIT, if you have a really important password
+          to hash and don't mind using 1GB of ram and 10s+ of CPU time on an i7-class CPU, you can use 
+          crypto_pwhash_OPSLIMIT_SENSITIVE and crypto_pwhash_MEMLIMIT_SENSITIVE instead.
+
+=item crypto_scalarmult_base()
+   
+   Usage: my $pk = crypto_scalarmult_base($sk);
+
+=item crypto_scalarmult()
+   
+   Usage: my $shared_secret = crypto_scalarmult($alice_secret, $bob_public);
+   
+=item crypto_scalarmult_safe()
+
+   Usage: my $shared_secret = crypto_scalarmult_safe($alice_secret, $bob_public, $alice_public);
+   
+   Note:  The shared secret generated is a hash of the output of crypto_scalarmult xor'd with the two public 
+          keys as outlined here L<https://download.libsodium.org/doc/advanced/scalar_multiplication.html>.
 
 =back
 
@@ -442,6 +514,8 @@ Crypt::Sodium - Perl bindings for libsodium (NaCL) https://github.com/jedisct1/l
  crypto_box_MACBYTES
  crypto_box_SEEDBYTES
  crypto_secretbox_MACBYTES
+ crypto_secretbox_KEYBYTES
+ crypto_secretbox_NONCEBYTES
  crypto_sign_PUBLICKEYBYTES
  crypto_sign_SECRETKEYBYTES
  crypto_pwhash_SALTBYTES
